
from __future__ import annotations

from dataclasses import dataclass
from json import dumps
from typing import (
    Any,
    Dict,
    List,
    Optional,
)
from urllib.parse import parse_qs, urlparse

from bamboo.base import HTTPMethods, MediaTypes
from bamboo.util.convert import unparse_qs


__all__ = []


@dataclass
class HTTPRequestForm:

    host: str
    port: Optional[int]
    uri: str
    method: str
    headers: Dict[str, str]
    body: Optional[bytes]


def get_http_request_form(
    scheme: str,
    uri: str,
    method: str,
    headers: Dict[str, str] = {},
    body: Optional[bytes] = None,
    json: Optional[Dict[str, Any]] = None,
    query: Dict[str, List[str]] = {}
) -> HTTPRequestForm:
    # method management
    method = method.upper()
    if method not in HTTPMethods:
        raise ValueError(f"Specified method '{method}' is not available.")

    # body management
    if body and json:
        raise ValueError("Request body is specified both 'body' and 'json'.")
    if body:
        if "Content-Type" not in headers:
            headers["Content-Type"] = MediaTypes.plain
    if json:
        body = dumps(json)
        if "Content-Type" not in headers:
            headers["Content-Type"] = MediaTypes.json

    parsed_uri = urlparse(uri)
    if parsed_uri.scheme != scheme:
        raise ValueError(
            f"Scheme of specified uri '{parsed_uri.scheme}' is "
            "not available. Use HTTP."
        )

    # port management
    port = parsed_uri.port
    if not port:
        port = None

    # query management
    query_included = parse_qs(parsed_uri.query)
    query_included.update(query)
    query = unparse_qs(query_included)

    # path management
    path = parsed_uri.path
    if len(query):
        path = "?".join((path, query))

    return HTTPRequestForm(
        parsed_uri.hostname,
        port,
        path,
        method,
        headers,
        body
    )



import unittest

from bamboo import (
    WSGIApp,
    WSGIEndpoint,
    VersionConfig,
)


app = WSGIApp()


@app.route("test", "hoge", version=1)
class TestEndpointSingle(WSGIEndpoint):
    ideal_uris = [
        ("v1", "test", "hoge"),
    ]


@app.route("test", "hogehoge", version=(1, 2, 3))
class TestEndpointMultiple(WSGIEndpoint):
    ideal_uris = [
        ("v1", "test", "hogehoge"),
        ("v2", "test", "hogehoge"),
        ("v3", "test", "hogehoge"),
    ]


@app.route("test", "hogest", version=None)
class TestEndpointAny(WSGIEndpoint):
    ideal_uris = [
        ("test", "hogest")
    ]


class TestEndpointNothing(WSGIEndpoint):
    pass


class TestVersion(unittest.TestCase):

    def test_version_single(self):
        config = VersionConfig(TestEndpointSingle)
        self.assertEqual(config.get(app), (1,))

        uris = app.search_uris(TestEndpointSingle)
        for uri, ideal in zip(uris, TestEndpointSingle.ideal_uris):
            self.assertEqual(uri, ideal)

    def test_version_multiple(self):
        config = VersionConfig(TestEndpointMultiple)
        self.assertEqual(config.get(app), (1, 2, 3))

        uris = app.search_uris(TestEndpointMultiple)
        for uri, ideal in zip(uris, TestEndpointMultiple.ideal_uris):
            self.assertEqual(uri, ideal)

    def test_version_any(self):
        config = VersionConfig(TestEndpointAny)
        self.assertEqual(config.get(app), ())

        uris = app.search_uris(TestEndpointAny)
        for uri, ideal in zip(uris, TestEndpointAny.ideal_uris):
            self.assertEqual(uri, ideal)

    def test_version_nothing(self):
        config = VersionConfig(TestEndpointNothing)
        self.assertEqual(config.get(app), None)


if __name__ == "__main__":
    unittest.main()



import unittest

from bamboo import (
    DEFAULT_CONTENT_TYPE_PLAIN,
    WSGIApp,
    WSGIEndpoint,
    WSGIServerForm,
    WSGITestExecutor,
)
from bamboo.request import http

from . import get_log_name


app = WSGIApp()
PATH_SERVER_LOG = get_log_name(__file__)


@app.route("hello")
class MockEndpoint(WSGIEndpoint):

    def do_GET(self) -> None:
        self.send_body(b"Hello, World!")


@app.route("attributes")
class AttributesEndpoint(WSGIEndpoint):

    def do_GET(self) -> None:
        assert self.wsgi_version == "1.0"
        assert self.server_software == "WSGIServer/0.2"
        assert self.http_version == "1.1"
        assert self.scheme == "http"
        assert self.get_host_addr() == ("localhost", 8000)
        assert self.path == "/attributes"
        assert self.queries == {}
        assert self.content_type == DEFAULT_CONTENT_TYPE_PLAIN
        assert self.method == "GET"

        self.send_only_status()


@app.route("callback")
class CallbackEndpoint(WSGIEndpoint):

    def pre_GET(self) -> None:
        self.word = "word"

    def do_GET(self) -> None:
        assert self.word == "word"
        self.send_only_status()


class WSGIAppTest(unittest.TestCase):

    @classmethod
    def setUpClass(cls) -> None:
        form = WSGIServerForm("", 8000, app, PATH_SERVER_LOG)
        cls.executor = WSGITestExecutor(form).start_serve()

    @classmethod
    def tearDownClass(cls) -> None:
        cls.executor.close()

    def test_hello(self):
        with http.get("http://localhost:8000/hello") as res:
            self.assertEqual(res.body, b"Hello, World!")

    def test_attributes(self):
        with http.get("http://localhost:8000/attributes") as res:
            self.assertTrue(res.ok)

    def test_callback(self):
        with http.get("http://localhost:8000/callback") as res:
            self.assertTrue(res.ok)


if __name__ == "__main__":
    unittest.main()




from typing import (
    Any,
    Awaitable,
    Callable,
    Tuple,
    Union,
)

from bamboo.endpoint import ASGIHTTPEndpoint, WSGIEndpoint


__all__ = []


# Signature of the callback of response method on sub classes of
# the Endpoint class.
Callback_WSGI_t = Callable[[WSGIEndpoint, Tuple[Any, ...]], None]
Callback_ASGI_t = Callable[[ASGIHTTPEndpoint, Tuple[Any, ...]], Awaitable[None]]
Callback_t = Union[Callback_WSGI_t, Callback_ASGI_t]
CallbackDecorator_t = Callable[[Callback_t], Callback_t]


def _get_bamboo_attr(attr: str) -> str:
    return f"__bamboo_{attr}__"



from collections import deque
from io import BytesIO, IOBase
import os
from typing import (
    Deque,
    Iterable,
    Iterator,
    TypeVar,
    Union,
)


__all__ = []


class ResponseBodyIteratorBase:
    """Base class of iterator for buffering response body.
    """

    def __iter__(self) -> Iterable[bytes]:
        return self

    def __next__(self) -> bytes:
        pass

    def __del__(self) -> None:
        pass


class BufferedBinaryIterator(ResponseBodyIteratorBase):
    """Iterator for sending bytes as chunks.

    Its object can be used to generate iterator from a bytes object, which
    yields specified size of bytes objects.

    Note:
        Even if you generate its object from a bytes object sending to a client
        and specify it as response body, there may be no effect on performance
        of your web application. Its object should be used only for unifying
        interface as a subclass of `ResponseBodyIteratorBase`.
    """

    def __init__(self, data: bytes, bufsize: int = 8192) -> None:
        """
        Args:
            data: A bytes object to be iterated.
            bufsize: Chunk size of bytes objects yielded from the iterator.
        """
        super().__init__()

        self._data = BytesIO(data)
        self._bufsize = bufsize

    def __next__(self) -> bytes:
        buffer = self._data.read(self._bufsize)
        if buffer:
            return buffer
        raise StopIteration()

    def __del__(self) -> None:
        self._data.close()


BinaryReadableStream_t = TypeVar("BinaryReadableStream_t")


class BufferedStreamIterator(ResponseBodyIteratorBase):
    """Iterator of binary made with file-like object.

    Its object can be used to iterate file-like obj of binary.
    """

    def __init__(
        self,
        stream: BinaryReadableStream_t,
        bufsize: int = 8192
    ) -> None:
        """
        Args:
            stream: File-like object readable binary.
            bufsize: Chunk size of bytes objects yielded from the iterator.

        Raises:
            TypeError: Raised if `stream` is not readable.
        """
        super().__init__()

        if not (isinstance(stream, IOBase) and hasattr(stream, "read")):
            raise TypeError(
                "'stream' must be IOBase and have the 'read' method."
            )

        self._stream = stream
        self._bufsize = bufsize

    def __next__(self) -> bytes:
        buffer = self._stream.read(self._bufsize)
        if buffer:
            return buffer
        raise StopIteration()

    def __del__(self) -> None:
        self._stream.close()


class BufferedFileIterator(BufferedStreamIterator):
    """Iterator of binary made with file path.

    Its object can be used to iterate file-like obj made with file path.
    """

    def __init__(
        self,
        path: str,
        bufsize: int = 8192,
        remove: bool = False,
    ) -> None:
        """
        Args:
            path: File path.
            bufsize: Chunk size of bytes objects yielded from the iterator.
        """
        self._path = path
        self._remove = remove

        file = open(path, "br")
        super().__init__(file, bufsize=bufsize)

    def __del__(self) -> None:
        super().__del__()

        if self._remove:
            os.remove(self._path)


class BufferedIteratorWrapper(ResponseBodyIteratorBase):
    """Iterator wrapping iterator yeilding bytes objects.

    Its object holds an iterator inner, which can yeilds bytes objects
    with arbitary sizes, and its object yeilds bytes objects with specified
    chunk size, i.e. this object changes chunk size of bytes yielded from
    given iterator.
    """

    def __init__(self, iter: Iterator[bytes], bufsize: int = 8192) -> None:
        """
        Args:
            iter: Iterator wrapped.
            bufsize: Chunk size of bytes objects yielded from the iterator.
        """
        super().__init__()

        self._iter = iter
        self._bufsize = bufsize
        self._buffer = BytesIO()

    @property
    def _is_buffer_filled(self) -> bool:
        return self._buffer.tell() >= self._bufsize

    def _update_buffer(self) -> bytes:
        self._buffer.seek(0)
        data = self._buffer.read(self._bufsize)
        other = self._buffer.read()
        self._buffer.close()
        self._buffer = BytesIO(other)
        return data

    def __next__(self) -> bytes:
        if self._is_buffer_filled:
            return self._update_buffer()

        while True:
            try:
                chunk = next(self._iter)
            except StopIteration:
                if self._buffer.tell():
                    return self._update_buffer()
                else:
                    raise StopIteration()
            else:
                diff = self._bufsize - self._buffer.tell()
                chunk_size = len(chunk)
                if chunk_size >= diff:
                    self._buffer.write(chunk[:diff])
                    self._buffer.flush()
                    data = self._update_buffer()

                    if chunk_size > diff:
                        self._buffer.write(chunk[diff:])

                    return data
                else:
                    self._buffer.write(chunk)
                    continue

    def __del__(self) -> None:
        self._buffer.close()


class BufferedConcatIterator(ResponseBodyIteratorBase):
    """Iterator conbining serveral iterator yeilding bytes objects.

    Its object can concatenate several iterator or bytes objects and behaves
    as a single iterator yieding bytes objects with specified chunk size.
    """

    def __init__(
        self,
        *items: Union[bytes, Iterator[bytes]],
        bufsize: int = 8192
    ) -> None:
        """
        Args:
            *items: bytes objects or iterators yielding bytes.
            bufsize: Chunk size of bytes objects yielded from the iterator.
        """
        super().__init__()

        self._iters: Deque[Iterator[bytes]] = deque()
        self._current: Iterator[bytes] = None
        self._bufsize = bufsize
        self._buffer = BytesIO()

        for item in items:
            self.append(item)

    def append(self, item: Union[bytes, Iterator[bytes]]) -> None:
        """Add bytes or iterator yielding bytes inside.

        Args:
            item: bytes object or iterator yielding bytes.
        """
        if isinstance(item, bytes):
            self._iters.append(
                BufferedBinaryIterator(item, bufsize=self._bufsize)
            )
        else:
            self._iters.append(
                BufferedIteratorWrapper(item, bufsize=self._bufsize)
            )

    @property
    def _is_buffer_filled(self) -> bool:
        return self._buffer.tell() >= self._bufsize

    def _update_buffer(self) -> bytes:
        data = self._buffer.getvalue()
        self._buffer = BytesIO()
        return data

    def __next__(self) -> bytes:
        if self._current is None:
            if len(self._iters):
                self._current = self._iters.popleft()
            else:
                raise StopIteration()

        while True:
            try:
                chunk = next(self._current)
            except StopIteration:
                if len(self._iters):
                    self._current = self._iters.popleft()
                    continue
                else:
                    if self._buffer.tell():
                        return self._update_buffer()
                    else:
                        raise StopIteration()
            else:
                diff = self._bufsize - self._buffer.tell()
                chunk_size = len(chunk)
                if chunk_size >= diff:
                    self._buffer.write(chunk[:diff])
                    self._buffer.flush()
                    data = self._update_buffer()

                    if chunk_size > diff:
                        self._buffer.write(chunk[diff:])

                    return data
                else:
                    self._buffer.write(chunk)
                    continue

    def __del__(self) -> None:
        self._buffer.close()



from abc import ABCMeta, abstractmethod
from dataclasses import dataclass
import inspect
from typing import (
    Any,
    Dict,
    Optional,
    Set,
    Tuple,
    Type,
)

from bamboo.api import ApiData, ValidationFailedError
from bamboo.base import AuthSchemes
from bamboo.endpoint import ASGIHTTPEndpoint, WSGIEndpoint
from bamboo.error import (
    DEFAULT_BASIC_AUTH_HEADER_NOT_FOUND_ERROR,
    DEFAULT_BEARER_AUTH_HEADER_NOT_FOUND_ERROR,
    DEFAULT_HEADER_NOT_FOUND_ERROR,
    DEFUALT_INCORRECT_DATA_FORMAT_ERROR,
    DEFAULT_NOT_APPLICABLE_IP_ERROR,
    ErrInfo,
)
from bamboo.sticky import (
    Callback_ASGI_t,
    Callback_WSGI_t,
    Callback_t,
    CallbackDecorator_t,
    _get_bamboo_attr,
)
from bamboo.util.convert import decode2binary
from bamboo.util.ip import is_valid_ipv4


__all__ = [
    "basic_auth",
    "bearer_auth",
    "data_format",
    "has_header_of",
    "has_query_of",
    "may_occur",
    "restricts_client",
]


class ConfigBase(metaclass=ABCMeta):

    ATTR: str

    @abstractmethod
    def get(self) -> Any:
        pass

    @abstractmethod
    def set(self, *args, **kwargs) -> Callback_t:
        pass


class HTTPErrorConfig(ConfigBase):

    ATTR = _get_bamboo_attr("errors")

    def __init__(self, callback) -> None:
        if not hasattr(callback, self.ATTR):
            setattr(callback, self.ATTR, set())

        self._callback = callback
        self._registered: Set[ErrInfo] = getattr(callback, self.ATTR)

    def get(self) -> Tuple[Type[ErrInfo], ...]:
        return tuple(self._registered)

    def set(self, *errors: Type[ErrInfo]) -> Callback_t:
        for err in errors:
            self._registered.add(err)
        return self._callback


def may_occur(*errors: Type[ErrInfo]) -> CallbackDecorator_t:
    """Register error classes to callback on `Endpoint`.

    Args:
        *errors: Error classes which may occuur.

    Returns:
        CallbackDecorator_t: Decorator to register
            error classes to callback.

    Examples:
        ```python
        class MockErrInfo(ErrInfo):
            http_status = HTTPStatus.INTERNAL_SERVER_ERROR

            def get_body(self) -> bytes:
                return b"Intrernal server error occured"

        class MockEndpoint(WSGIEndpoint):

            @may_occur(MockErrInfo)
            def do_GET(self) -> None:
                # Do something...

                # It is possible to send error response.
                if is_some_flag():
                    self.send_err(MockErrInfo())

                self.send_only_status()
        ```
    """

    def wrapper(callback: Callback_t) -> Callback_t:
        config = HTTPErrorConfig(callback)
        return config.set(*errors)

    return wrapper


@dataclass(eq=True, frozen=True)
class DataFormatInfo:
    """`dataclass` with information of data format at callbacks on `Endpoint`.

    Attributes:
        input (Optional[Type[ApiData]]): Input data format.
        output (Optional[Type[ApiData]]): Output data format.
        is_validate (bool): If input data is to be validate.
        err_validate (ErrInfo): Error information sent
            when validation failes.
    """
    input: Optional[Type[ApiData]] = None
    output: Optional[Type[ApiData]] = None
    is_validate: bool = True
    err_validate: ErrInfo = DEFUALT_INCORRECT_DATA_FORMAT_ERROR
    err_noheader: ErrInfo = DEFAULT_HEADER_NOT_FOUND_ERROR


class DataFormatConfig(ConfigBase):

    ATTR = _get_bamboo_attr("data_format")

    def __init__(self, callback: Callback_t) -> None:
        super().__init__()

        self._callback = callback

    def get(self) -> Optional[DataFormatInfo]:
        return getattr(self._callback, self.ATTR, None)

    def set(self, dataformat: DataFormatInfo) -> Callback_t:
        setattr(self._callback, self.ATTR, dataformat)
        if not dataformat.is_validate:
            return self._callback

        if inspect.iscoroutinefunction(self._callback):
            if dataformat.input:
                func = self.decorate_asgi
            else:
                func = self.decorate_asgi_no_input
        else:
            if dataformat.input:
                func = self.decorate_wsgi
            else:
                func = self.decorate_wsgi_no_input

        return func(self._callback, dataformat)

    @staticmethod
    def decorate_wsgi(
        callback: Callback_WSGI_t,
        dataformat: DataFormatInfo
    ) -> Callback_WSGI_t:

        @may_occur(dataformat.err_validate.__class__)
        @has_header_of("Content-Type", dataformat.err_noheader, add_arg=False)
        def _callback(self: WSGIEndpoint, *args) -> None:
            body = self.body
            try:
                data = dataformat.input(body, self.content_type)
            except ValidationFailedError:
                raise dataformat.err_validate

            callback(self, data, *args)

        _callback.__dict__ = callback.__dict__
        return _callback

    @staticmethod
    def decorate_wsgi_no_input(
        callback: Callback_WSGI_t,
        dataformat: DataFormatInfo
    ) -> Callback_WSGI_t:

        def _callback(self: WSGIEndpoint, *args) -> None:
            setattr(self, "body", b"")
            callback(self, *args)

        _callback.__dict__ = callback.__dict__
        return _callback

    @staticmethod
    def decorate_asgi(
        callback: Callback_ASGI_t,
        dataformat: DataFormatInfo
    ) -> Callback_ASGI_t:

        @may_occur(dataformat.err_validate.__class__)
        @has_header_of("Content-Type", dataformat.err_noheader, add_arg=False)
        async def _callback(self: ASGIHTTPEndpoint, *args) -> None:
            body = await self.body
            try:
                data = dataformat.input(body, self.content_type)
            except ValidationFailedError:
                raise dataformat.err_validate

            await callback(self, data, *args)

        _callback.__dict__ = callback.__dict__
        return _callback

    @staticmethod
    def decorate_asgi_no_input(
        callback: Callback_ASGI_t,
        dataformat: DataFormatInfo
    ) -> Callback_ASGI_t:

        async def _callback(self: ASGIHTTPEndpoint, *args) -> None:
            # NOTE
            #   Reference awaitable_cached_property
            body_property = await self.__class__.body
            body_property._obj2val[self] = b""
            await callback(self, *args)

        _callback.__dict__ = callback.__dict__
        return _callback


def data_format(
    input: Optional[Type[ApiData]] = None,
    output: Optional[Type[ApiData]] = None,
    is_validate: bool = True,
    err_validate: ErrInfo = DEFUALT_INCORRECT_DATA_FORMAT_ERROR,
    err_noheader: ErrInfo = DEFAULT_HEADER_NOT_FOUND_ERROR,
) -> CallbackDecorator_t:
    """Set data format of input/output data as API to callback on
    `Endpoint`.

    This decorator can be used to add attributes of data format information
    to a response method, and execute validation if input raw data has
    expected format defined on `input` argument.

    To represent no data inputs/outputs, specify `input`/`output` arguments
    as `None`. If `input` is `None`, then any data received from client will
    not be read. If `is_validate` is `False`, then validation will not be
    executed.

    Args:
        input: Input data format.
        output: Output data format.
        is_validate: If input data is to be validated.
        err_validate: Error information sent when validation failes.

    Returns:
        Decorator to add attributes of data format information to callback.

    Examples:
        ```python
        class UserData(JsonApiData):
            name: str
            email: str
            age: int

        class MockEndpoint(WSGIEndpoint):

            @data_format(input=UserData, output=None)
            def do_GET(self, rec_body: UserData) -> None:
                user_name = rec_body.name
                # Do something...
        ```
    """
    dataformat = DataFormatInfo(
        input,
        output,
        is_validate,
        err_validate,
        err_noheader,
    )

    def wrapper(callback: Callback_t) -> Callback_t:
        config = DataFormatConfig(callback)
        return config.set(dataformat)

    return wrapper


@dataclass(eq=True, frozen=True)
class RequiredHeaderInfo:
    """`dataclass` with information of header which should be included in
    response headers.

    Attributes:
        header: Name of header.
        err: Error information sent when the header is not included.
    """
    header: str
    err: ErrInfo
    add_arg: bool


class RequiredHeaderConfig(ConfigBase):

    ATTR = _get_bamboo_attr("required_headers")

    def __init__(self, callback: Callback_t) -> None:
        super().__init__()

        if not hasattr(callback, self.ATTR):
            setattr(callback, self.ATTR, set())

        self._callback = callback
        self._registered: Set[RequiredHeaderInfo] = getattr(callback, self.ATTR)

    def get(self) -> Tuple[RequiredHeaderInfo]:
        return tuple(self._registered)

    def set(self, info: RequiredHeaderInfo) -> Callback_t:
        self._registered.add(info)

        if inspect.iscoroutinefunction(self._callback):
            func = self.decorate_asgi
        else:
            func = self.decorate_wsgi

        return func(self._callback, info)

    @staticmethod
    def decorate_wsgi(
        callback: Callback_WSGI_t,
        info: RequiredHeaderInfo
    ) -> Callback_WSGI_t:

        @may_occur(info.err.__class__)
        def _callback(self: WSGIEndpoint, *args) -> None:
            val = self.get_header(info.header)
            if val is None:
                raise info.err

            if info.add_arg:
                callback(self, val, *args)
            else:
                callback(self, *args)

        _callback.__dict__ = callback.__dict__
        return _callback

    @staticmethod
    def decorate_asgi(
        callback: Callback_ASGI_t,
        info: RequiredHeaderInfo
    ) -> Callback_ASGI_t:

        @may_occur(info.err.__class__)
        async def _callback(self: ASGIHTTPEndpoint, *args) -> None:
            val = self.get_header(info.header)
            if val is None:
                raise info.err

            if info.add_arg:
                await callback(self, val, *args)
            else:
                await callback(self, *args)

        _callback.__dict__ = callback.__dict__
        return _callback


def has_header_of(
    header: str,
    err: ErrInfo = DEFAULT_HEADER_NOT_FOUND_ERROR,
    add_arg: bool = True,
) -> CallbackDecorator_t:
    """Set callback up to receive given header from clients.

    If request headers don't include specified `header`, then response
    headers and body will be made based on `err` and sent.

    Args:
        header: Name of header.
        err: Error information sent when specified `header` is not found.

    Returns:
        Decorator to make callback to be set up to receive the header.

    Examples:
        ```python
        class BasicAuthHeaderNotFoundErrInfo(ErrInfo):
            http_status = HTTPStatus.UNAUTHORIZED

            def get_headers(self) -> List[Tuple[str, str]]:
                return [("WWW-Authenticate", 'Basic realm="SECRET AREA"')]

        class MockEndpoint(WSGIEndpoint):

            @has_header_of("Authorization", BasicAuthHeaderNotFoundErrInfo())
            def do_GET(self) -> None:
                # It is guaranteed that request headers include the
                # `Authorization` header at this point.
                header_auth = self.get_header("Authorization")

                # Do something...
        ```
    """
    info = RequiredHeaderInfo(header, err, add_arg)

    def wrapper(callback: Callback_t) -> Callback_t:
        config = RequiredHeaderConfig(callback)
        return config.set(info)

    return wrapper


@dataclass(eq=True, frozen=True)
class ClientInfo:

    ip: str
    port: Optional[int] = None


_RestrictedClient_t = Dict[str, Set[Optional[int]]]


class RestrictedClientsConfig(ConfigBase):

    ATTR = _get_bamboo_attr("restricted_clients")

    def __init__(self, callback: Callback_t) -> None:
        super().__init__()

        if not hasattr(callback, self.ATTR):
            setattr(callback, self.ATTR, {})

        self._callback = callback
        self._registered: _RestrictedClient_t = getattr(callback, self.ATTR)

    def get(self) -> _RestrictedClient_t:
        return self._registered.copy()

    def set(
        self,
        *clients: ClientInfo,
        err: ErrInfo = DEFAULT_NOT_APPLICABLE_IP_ERROR
    ) -> Callback_t:
        for client in clients:
            if not is_valid_ipv4(client.ip):
                raise ValueError(f"{client.ip} is an invalid IP address.")
            if client.ip == "localhost":
                client = ClientInfo("127.0.0.1", client.port)

            ports = self._registered.get(client.ip)
            if ports is None:
                self._registered[client.ip] = set()
            self._registered[client.ip].add(client.port)

        if inspect.iscoroutinefunction(self._callback):
            func = self.decorate_asgi
        else:
            func = self.decorate_wsgi
        return func(self._callback, err)

    @classmethod
    def decorate_wsgi(
        cls,
        callback: Callback_WSGI_t,
        err: ErrInfo
    ) -> Callback_WSGI_t:
        acceptables = getattr(callback, cls.ATTR, {})

        @may_occur(err.__class__)
        def _callback(self: WSGIEndpoint, *args) -> None:
            client = ClientInfo(*self.get_client_addr())
            ports = acceptables.get(client.ip)
            if ports is None:
                raise err
            if not(None in ports or client.port in ports):
                raise err

            callback(self, *args)

        _callback.__dict__ = callback.__dict__
        return _callback

    @classmethod
    def decorate_asgi(
        cls,
        callback: Callback_ASGI_t,
        err: ErrInfo
    ) -> Callback_ASGI_t:
        acceptables = getattr(callback, cls.ATTR, set())

        @may_occur(err.__class__)
        async def _callback(self: ASGIHTTPEndpoint, *args) -> None:
            client = ClientInfo(*self.get_client_addr())
            ports = acceptables.get(client.ip)
            if ports is None:
                raise err
            if not(None in ports or client.port in ports):
                raise err

            await callback(self, *args)

        _callback.__dict__ = callback.__dict__
        return _callback


def restricts_client(
    *clients: ClientInfo,
    err: ErrInfo = DEFAULT_NOT_APPLICABLE_IP_ERROR
) -> CallbackDecorator_t:
    """Restrict IP addresses at callback.

    Args:
        *client_ips: IP addresses to be allowed to request
        err: Error information sent when request from IP not included
            specified IPs comes.

    Returns:
        Decorator to make callback to be set up to restrict IP addresses.

    Raises:
        ValueError: Raised if invalid IP address is detected

    Examples:
        ```python
        class MockEndpoint(WSGIEndpoint):

            # Restrict to allow only localhost to request
            # to this callback
            @restricts_client(ClientInfo("localhost"))
            def do_GET(self) -> None:
                # Only localhost can access to the callback.

                # Do something...
        ```
    """
    def wrapper(callback: Callback_t) -> Callback_t:
        config = RestrictedClientsConfig(callback)
        return config.set(*clients, err=err)

    return wrapper


class MultipleAuthSchemeError(Exception):
    """Raised if several authentication schemes of the 'Authorization'
    header are detected.
    """
    pass


class AuthSchemeConfig(ConfigBase):

    ATTR = _get_bamboo_attr("auth_scheme")
    HEADER_AUTHORIZATION = "Authorization"

    def __init__(self, callback: Callback_t) -> None:
        super().__init__()

        self._callback = callback

    def get(self) -> Optional[str]:
        return getattr(self._callback, self.ATTR, None)

    def set(self, scheme: str, err: ErrInfo) -> Callback_t:
        if hasattr(self._callback, self.ATTR):
            _scheme_registered = getattr(self._callback, self.ATTR)
            raise MultipleAuthSchemeError(
                "Authentication scheme has already been specified as "
                f"'{_scheme_registered}'. Do not specify multiple schemes."
            )

        if scheme not in AuthSchemes:
            raise ValueError(f"Specified scheme '{scheme}' is not supported.")

        setattr(self._callback, self.ATTR, scheme)

        if scheme == AuthSchemes.basic:
            if inspect.iscoroutinefunction(self._callback):
                func = self.decorate_asgi_basic
            else:
                func = self.decorate_wsgi_basic
        elif scheme == AuthSchemes.bearer:
            if inspect.iscoroutinefunction(self._callback):
                func = self.decorate_asgi_bearer
            else:
                func = self.decorate_wsgi_bearer

        return func(self._callback, err)

    @staticmethod
    def _validate_auth_header(value: str, scheme: str) -> Optional[str]:
        value = value.split(" ")
        if len(value) != 2:
            return None

        _scheme, credentials = value
        if _scheme != scheme:
            return None

        return credentials

    @classmethod
    def decorate_wsgi_basic(
        cls,
        callback: Callback_WSGI_t,
        err: ErrInfo
    ) -> Callback_WSGI_t:

        @may_occur(err.__class__)
        @has_header_of(cls.HEADER_AUTHORIZATION, err, add_arg=False)
        def _callback(self: WSGIEndpoint, *args) -> None:
            val = self.get_header(cls.HEADER_AUTHORIZATION)
            credentials = cls._validate_auth_header(val, AuthSchemes.basic)
            if credentials is None:
                raise err

            credentials = decode2binary(credentials).decode().split(":")
            if len(credentials) != 2:
                raise err

            user_id, pw = credentials
            callback(self, user_id, pw, *args)

        _callback.__dict__ = callback.__dict__
        return _callback

    @classmethod
    def decorate_asgi_basic(
        cls,
        callback: Callback_ASGI_t,
        err: ErrInfo
    ) -> Callback_ASGI_t:

        @may_occur(err.__class__)
        @has_header_of(cls.HEADER_AUTHORIZATION, err, add_arg=False)
        async def _callback(self: ASGIHTTPEndpoint, *args) -> None:
            val = self.get_header(cls.HEADER_AUTHORIZATION)
            credentials = cls._validate_auth_header(val, AuthSchemes.basic)
            if credentials is None:
                raise err

            credentials = decode2binary(credentials).decode().split(":")
            if len(credentials) != 2:
                raise err

            user_id, pw = credentials
            await callback(self, user_id, pw, *args)

        _callback.__dict__ = callback.__dict__
        return _callback

    @classmethod
    def decorate_wsgi_bearer(
        cls,
        callback: Callback_WSGI_t,
        err: ErrInfo,
    ) -> Callback_WSGI_t:

        @may_occur(err.__class__)
        @has_header_of(cls.HEADER_AUTHORIZATION, err, add_arg=False)
        def _callback(self: WSGIEndpoint, *args) -> None:
            val = self.get_header(cls.HEADER_AUTHORIZATION)
            token = cls._validate_auth_header(val, AuthSchemes.bearer)
            if token is None:
                raise err

            callback(self, token, *args)

        _callback.__dict__ = callback.__dict__
        return _callback

    @classmethod
    def decorate_asgi_bearer(
        cls,
        callback: Callback_ASGI_t,
        err: ErrInfo,
    ) -> Callback_ASGI_t:

        @may_occur(err.__class__)
        @has_header_of(cls.HEADER_AUTHORIZATION, err, add_arg=False)
        async def _callback(self: ASGIHTTPEndpoint, *args) -> None:
            val = self.get_header(cls.HEADER_AUTHORIZATION)
            token = cls._validate_auth_header(val, AuthSchemes.bearer)
            if token is None:
                raise err

            await callback(self, token, *args)

        _callback.__dict__ = callback.__dict__
        return _callback


def basic_auth(
    err: ErrInfo = DEFAULT_BASIC_AUTH_HEADER_NOT_FOUND_ERROR,
) -> CallbackDecorator_t:
    """Set callback up to require `Authorization` header in Basic
    authentication.

    Args:
        err: Error sent when `Authorization` header is not found, received
            scheme doesn't match, or extracting user ID and password from
            credentials failes.

    Returns:
        Decorator to make callback to be set  up to require
        the `Authorization` header.

    Examples:
        ```python
        class MockEndpoint(WSGIEndpoint):

            @basic_auth()
            def do_GET(self, user_id: str, password: str) -> None:
                # It is guaranteed that request headers include the
                # `Authorization` header at this point, and user_id and
                # password are the ones extracted from the header.

                # Authenticate with any function working on your system.
                authenticate(user_id, password)

                # Do something...
        ```
    """
    def wrapper(callback: Callback_t) -> Callback_t:
        config = AuthSchemeConfig(callback)
        return config.set(AuthSchemes.basic, err)

    return wrapper


def bearer_auth(
    err: ErrInfo = DEFAULT_BEARER_AUTH_HEADER_NOT_FOUND_ERROR,
) -> CallbackDecorator_t:
    """Set callback up to require `Authorization` header in token
    authentication for OAuth 2.0.

    Args:
        err : Error sent when `Authorization` header is not found, or
            when received scheme doesn't match.

    Returns:
        Decorator to make callback to be set  up to require
        the `Authorization` header.

    Examples:
        ```python
        class MockEndpoint(WSGIEndpoint):

            @bearer_auth()
            def do_GET(self, token: str) -> None:
                # It is guaranteed that request headers include the
                # `Authorization` header at this point, and token is
                # the one extracted from the header.

                # Authenticate with any function working on your system.
                authenticate(token)

                # Do something...
        ```
    """
    def wrapper(callback: Callback_t) -> Callback_t:
        config = AuthSchemeConfig(callback)
        return config.set(AuthSchemes.bearer, err)

    return wrapper


@dataclass(eq=True, frozen=True)
class RequiredQueryInfo:

    query: str
    err_empty: Optional[ErrInfo] = None
    err_not_unique: Optional[ErrInfo] = None
    add_arg: bool = True


class RequiredQueryConfig(ConfigBase):

    ATTR = _get_bamboo_attr("required_queries")

    def __init__(self, callback: Callback_t) -> None:
        super().__init__()

        if not hasattr(callback, self.ATTR):
            setattr(callback, self.ATTR, set())

        self._callback = callback
        self._registered: Set[RequiredQueryConfig] = getattr(callback, self.ATTR)

    def get(self) -> Tuple[RequiredHeaderInfo]:
        return tuple(self._registered)

    def set(self, info: RequiredQueryInfo) -> Callback_t:
        self._registered.add(info)

        if inspect.iscoroutinefunction(self._callback):
            func = self.decorate_asgi
        else:
            func = self.decorate_wsgi
        return func(self._callback, info)

    @staticmethod
    def decorate_wsgi(
        callback: Callback_WSGI_t,
        info: RequiredQueryInfo,
    ) -> Callback_WSGI_t:

        def _callback(self: WSGIEndpoint, *args) -> None:
            val = self.get_queries(info.query)
            len_val = len(val)

            if len_val == 0:
                if info.err_empty:
                    raise info.err_empty
                else:
                    val = None
            elif len_val == 1:
                if info.err_not_unique:
                    raise info.err_not_unique
                else:
                    val = val[0]

            if info.add_arg:
                callback(self, val, *args)
            else:
                callback(self, *args)

        errs = []
        if info.err_empty:
            errs.append(info.err_empty.__class__)
        if info.err_not_unique:
            errs.append(info.err_not_unique.__class__)
        if len(errs):
            _callback = may_occur(*errs)(_callback)

        _callback.__dict__.update(callback.__dict__)
        return _callback

    @staticmethod
    def decorate_asgi(
        callback: Callback_ASGI_t,
        info: RequiredQueryInfo,
    ) -> Callback_ASGI_t:

        async def _callback(self: ASGIHTTPEndpoint, *args) -> None:
            val = self.get_queries(info.query)
            len_val = len(val)

            if len_val == 0:
                if info.err_empty:
                    raise info.err_empty
                else:
                    val = None
            elif len_val == 1:
                if info.err_not_unique:
                    raise info.err_not_unique
                else:
                    val = val[0]

            if info.add_arg:
                await callback(self, val, *args)
            else:
                await callback(self, *args)

        errs = []
        if info.err_empty:
            errs.append(info.err_empty.__class__)
        if info.err_not_unique:
            errs.append(info.err_not_unique.__class__)
        if len(errs):
            _callback = may_occur(*errs)(_callback)

        _callback.__dict__.update(callback.__dict__)
        return _callback


def has_query_of(
    query: str,
    err_empty: Optional[ErrInfo] = None,
    err_not_unique: Optional[ErrInfo] = None,
    add_arg: bool = True,
) -> CallbackDecorator_t:
    """Set callback up to receive given query parameter from clients.
    """
    info = RequiredQueryInfo(query, err_empty, err_not_unique, add_arg)

    def wrapper(callback: Callback_t) -> Callback_t:
        config = RequiredQueryConfig(callback)
        return config.set(info)

    return wrapper




import base64
from typing import (
    Dict,
    List,
    Union,
)


__all__ = [
    "decode2binary",
    "encode_binary",
    "encode_base64_string",
    "unparse_qs",
]


def encode_binary(data: Union[bytes, bytearray]) -> str:
    return base64.b64encode(data).decode()


def encode_base64_string(data: str) -> str:
    return encode_binary(data.encode())


def decode2binary(data: str) -> bytes:
    return base64.b64decode(data.encode())


def unparse_qs(query: Dict[str, List[str]]) -> str:
    params = []
    for key, vals in query.items():
        repr_vals = ",".join(vals)
        params.append("=".join((key, repr_vals)))
    return "&".join(params)



from os import getrandom
from typing import Generator
import unittest

from bamboo import BufferedIteratorWrapper


def test_generator() -> Generator[bytes, None, None]:
    for i in range(10**2, 10**3):
        yield getrandom(i)


class BufferedIteratorWrapperTest(unittest.TestCase):

    def setUp(self) -> None:
        self.total = sum(range(10**2, 10**3))
        self.bufsize = 8192
        self.iter = BufferedIteratorWrapper(test_generator(), self.bufsize)

    def test_iter(self):
        sum = 0
        for i in self.iter:
            diff = self.total - sum
            if diff < self.bufsize:
                self.assertEqual(len(i), diff)
            else:
                self.assertEqual(len(i), self.bufsize)

            sum += len(i)
        self.assertEqual(sum, self.total)


if __name__ == "__main__":
    unittest.main()



from typing import (
    Dict,
    Generic,
    List,
    Optional,
    Tuple,
    Type,
    TypeVar, Union,
)

from bamboo.location import (
    FlexibleLocation,
    Uri_t,
    is_flexible_uri,
    is_duplicated_uri,
)


__all__ = []


HTTPMethod_t = str
Endpoint_t = TypeVar("Endpoint_t")
Uri2Endpoints_t = Dict[Uri_t, Type[Endpoint_t]]


class DuplicatedUriRegisteredError(Exception):
    """Raised if duplicated URI is registered."""
    pass


class Router(Generic[Endpoint_t]):
    """Operator of routing request to `Endpoint` by URI.
    """

    def __init__(self) -> None:
        self._raw_uri2endpoint: Uri2Endpoints_t = {}
        self.uri2endpoint: Uri2Endpoints_t = {}
        self.uris_flexible: List[Uri_t] = []

    def register(
        self,
        uri: Uri_t,
        endpoint: Type[Endpoint_t],
        version: Union[str, Tuple[str, ...]] = ()
    ) -> None:
        """Register combination of URI and `Endpoint`.

        Args:
            uri: URI pattern of the `Endpoint`.
            endpoint: `Endpoint` class to be registered.
            version: Version of the `Endpoint`.

        Raises:
            DuplicatedUriRegisteredError: Raised if given URI pattern
                matches one already registered.
        """
        for uri_registered in self._raw_uri2endpoint.keys():
            if is_duplicated_uri(uri_registered, uri):
                raise DuplicatedUriRegisteredError(
                    "Duplicated URIs were detected.\n"
                    f"URI pattern 1: {uri_registered}\n"
                    f"URI pattern 2: {uri}"
                )
        else:
            self._raw_uri2endpoint[uri] = endpoint

        if isinstance(version, str):
            version = (version,)

        if len(version):
            uris = [(ver,) + uri for ver in version]
        else:
            uris = [uri]

        for _uri in uris:
            if is_flexible_uri(_uri):
                self.uris_flexible.append(_uri)
            self.uri2endpoint[_uri] = endpoint

    def validate(
        self,
        uri: str
    ) -> Tuple[Tuple[str, ...], Optional[Type[Endpoint_t]]]:
        """Validate specified `uri` and retrieved `Endpoint`.

        Note:
            This method returns pair of tuple of flexible locations specified as
            parts of URI pattern linked to `Endpoint` and its `Endpoint`. If any
            flexible locations are not included in the URI pattern, then empty
            tuple will be returned as a sequence of flexible locations, so if
            URI patterns is configured with only static locations, you will get
            the empty tuple.

            If invalid URI pattern is come, then also empty tuple will be return
            as sequence of flexible locations and `None` as `Endpoint`, or
            `((), None)`.

        Args:
            uri: Path of URI.

        Returns:
            Pair of values of flexible locations and `Endpoint` if specified
            `uri` is valid.
        """
        uri = tuple(uri[1:].split("/"))
        if not uri[0]:
            uri = ()

        endpoint = self.uri2endpoint.get(uri)
        if endpoint:
            return ((), endpoint)

        depth = len(uri)
        for flexible in self.uris_flexible:
            if len(flexible) != depth:
                continue

            flexibles_received = []

            # Judging each locations
            for loc_req, loc_flex in zip(uri, flexible):
                if isinstance(loc_flex, FlexibleLocation):
                    if not loc_flex.is_valid(loc_req):
                        break

                    flexibles_received.append(loc_req)
                else:
                    if loc_req != loc_flex:
                        break
            else:
                # Correct case
                endpoint = self.uri2endpoint.get(flexible)
                return (tuple(flexibles_received), endpoint)

        # Could not find it
        return ((), None)

    def search_uris(self, endpoint: Type[Endpoint_t]) -> List[Uri_t]:
        """Search URI patterns of specified `endpoint`.

        Args:
        endpoint: `Endpoint` class whose URI patterns to be retrieved.

        Returns:
            Result of searching.
        """
        return [
            uri for uri, point in self.uri2endpoint.items()
            if point is endpoint
        ]




from datetime import datetime
from email.utils import formatdate, parsedate_to_datetime
from typing import Optional


__all__ = [
    "get_datetime_rfc822",
    "rfc822_to_datetime",
    "rfc822_to_unix",
    "unix_to_rfc822",
]


def get_datetime_rfc822(timeval: Optional[float] = None) -> str:
    return formatdate(timeval=timeval, usegmt=True)


def unix_to_rfc822(unixtime: float) -> str:
    return get_datetime_rfc822(unixtime)


def rfc822_to_datetime(rfc822: str) -> datetime:
    return parsedate_to_datetime(rfc822)


def rfc822_to_unix(rfc822: str) -> float:
    dt = rfc822_to_datetime(rfc822)
    return dt.timestamp()
